/*
 *******************************************************************
 *
 * FILE
 *
 * prover.java
 * by Phuong, The NGUYEN
 *
 * a prover
 * as part of a summer scholarship in RSISE - ANU
 * supervised by: Rajeev, Prabhakar GORE
 *
 * begin: 27 Jan 01
 * last updated: 21 Feb 01
 *
 *******************************************************************
 * 
 * WHAT IT DOES
 *
 * Given a parse tree of a formula in the logic S4, it will try to
 * satify that formula by searching for a model for it.
 * 
 * This class will reside in the Java smartcard
 *
 *******************************************************************
 *
 * HOW IT WORKS
 *
 * The program will perform a search for the model by searching for
 * the set of worlds that satisfy the set of formulae generated from
 * the original formula by the tableau rules.
 * The search is Depth Search First.
 *
 * Currently it assumes that the formula given to it is in NNF
 *
 *******************************************************************
 */


/* to reside in a card, it must belong to a package and implement a 
 * common interface so that the PC (client) can communicate with it 
 */
package card;
public class prover extends javacard.framework.Applet
	implements proverInterface
{
	/* all of the node in the parse tree will be numbered from 0 (root)
	 * The numbering is done so that the root is always 0, and nodes in
	 * lower level (larger depth) will be given larger index than that
	 * of nodes in higher levels.
	 * the nature list will indicate the nature of each node. Thus
	 * nature[0] = DISJ if the root is a disjunction ... etc.
	 */

	private byte _childs [], _nature [];
	byte _num_forms;

	/* since the communications with the card do not support
	 * transmissions of closed to 64 bytes, it is set up so that each
	 * transmission is for maximum of 35 bytes
	 */
	public void getChilds(byte []c, short start, byte len)
	{
		for(int i=0; i<len; i++) _childs[i+start] = c[i];
	}
	
	public void getNature(byte []n, byte start, byte len)
	{
		for(int i=0; i<len; i++) _nature[i+start] = n[i];
		_num_forms = (byte)(start + len);
	}


	/* constructor to be invoked when this prover is first loaded
	 * onto the card
	 */
	public prover ()
	{
		_childs = new byte[MAXFORMLEN << 1];
		_nature = new byte[MAXFORMLEN];
	}
	
	/* prove: the prove procedure.
	 *
	 * return 1 if the formula is satisfiable (if the root-node of the
	 * tableau is "good"), 0 if not.
	 *
	 * a logical child of a (non-saturated) tableau node is obtained by
	 * applying any of the non-K rules. A meta child of a (saturated)
	 * tableau node is obtained by applying one of the K-rules available
	 * to that tableau node
	 *
	 * a tableau node is "good" iff either
	 *		its core is the same as that of one of its ancestors
	 *		it has no childs, and its set of formulae contains no conflicts
	 *		it has logical childrens, and at least one of them is "good"
	 *		it has meta childrens, and all of them are "good"
	 *
	 * We need to check if the root-node is "good" or not
	 *
	 * ALGORITHM
	 *
	 * At each newly generated node (a meta child of some other node)
	 *		Check if it has appeared before in the branch
	 *		If yes: 
	 *			this node is good. Move up to its parent to
	 *			continue the cheking with its parent. this is done by
	 *			popping the top of stack (the parent node) and store it
	 *			in the cur_world, mark it with appropriate parameters.
	 *			Note if stack is empty, then shoud return 1. However, it
	 *			will never be empty in this case.
	 *		If not
	 *			Saturate it
	 *			Check if it contains conflict
	 *			If yes:
	 *				It is not good.
	 *				If there is no more sibling
	 *					its parent node is not good. Thus one needs to
	 *					pop the top of stack (the parent node), store
	 *					it in cur_world, and marks it with appropriate
	 *					paras (e.g isClosed = TRUE)
	 *					Note if stack is empty then return 0
	 *				If it possible to move to its next sibling
	 *					then move to next sibling
	 *			If not
	 *				If no more K-rule to be applied, this node is good
	 *				Move up to parent to continue the checking with it.
	 *				This is done by popping the top of stack (the parent
	 *				node) and store it in cur_world, mark it with
	 *				appropriate paras. Note that if stack is empty
	 *				then we should return 1
	 *
	 *				If there is more K-rules to be applied, pick the next
	 *				available, applied it to get tmp_world. Store 
	 *				cur_world in stack, update cur_world with tmp_world
	 *				set paras appropriately, and continue
	 *
	 *
	 * IMPLEMENTATIONS:
	 *
	 * We do not have a separate checklist, since it can be regerate from
	 * what are stored in stack. this reduces 50% of the space requirement
	 * Note that each tableau node requires n bits.
	 */

	public byte prove()
	{
		int i, j, k;
		byte num_forms = _num_forms;
		byte ROWLEN = (byte)((num_forms + 7) >> 3);

		byte childs[] = new byte[num_forms << 1];
		byte nature[] = new byte[num_forms];

		for(i=0; i<num_forms; i++)
		{
			j = i<<1;
			childs[j] = _childs[j];
			childs[j|1] = _childs[j|1];
			nature[i] = _nature[i];
		}


		ByteArray stack = null;
		byte current_lastPi	= num_forms;

		/* the current world is store at the cur_world array,
		 * Could use bits, but use byte for simplicity 
		 */
		byte cur_world[] = new byte [num_forms];
		byte tmp_world[] = new byte [num_forms];
		byte core[] = new byte [num_forms];

		byte isNewWorld = TRUE;
		byte isCheckedLoop = FALSE;
		byte isClosed = FALSE;
		byte b;
		
		/* current state that we are working on */
		cur_world[0] = 1;

		while(true)
		{
			/* if this is a newly generated world: cur_world will contain
			 * only its core and one of the child of a SOME formula. We
			 * need to check this *expanded* core against the other
			 * *expanded* cores in the stack
			 */
			if(isCheckedLoop == FALSE)
			{
				isNewWorld = TRUE;
				ByteArray stk = null;
				if(stack != null) stk = stack.next;
				for(; stk != null; stk = stk.next)
				{
					/* get the "expanded core" of the node at this position */
					for(i=0; i<num_forms; i++)
					{
						if(nature[i] == ALL)
						{ 
							b = (byte)(stk.array[(i>>3)] & (1 << (i&MASK)));
							if(b != 0)
							{
								core[i] = 1;
							}
							else core[i] = 0;
						}
						else core[i] = 0;
					}
					/* also the transitional rule to move to this world */
					core[childs[stk.array[ROWLEN] << 1]] = 1;

					/* checking for repetition */
					for(i=0; i<num_forms; i++)
					{
						if(core[i] == 1)
						{
							if(cur_world[i] == 0) break;
						}
						else if(cur_world[i] == 1) break;
					}
					if(i==num_forms) break;
				}


				/* if it appear at level i of stack */
				if(stk != null)
				{
					/* go back one level in the stack */
					for(i=0; i<num_forms; i++)
					{
						b = (byte)(stack.array[(i>>3)] & (1 << (i&MASK)));
						if(b != 0)
						{
							cur_world[i] = 1;
						}
						else cur_world[i] = 0;
					}

					/* last application of Pi rule of current world */
					current_lastPi = stack.array[ROWLEN];
					stack = stack.next;					
										
					isCheckedLoop = TRUE;
					continue;
				}
				
				else
				{
					isCheckedLoop = TRUE;
				}
			} // if isCheckedLoop


			/* for a new world, expand it and check if the current set of 
			 * formulae appears before. If yes then it is "good".
			 */
			if(isNewWorld == TRUE)
			{
				/* saturate it
				 * we fulfill it with non-K rules. The order to apply is
				 * from the root of parse-tree (formulae index increases)
				 * Note that if one child of a *DISJ* node (in parsetree)
				 * is kept due to the transitivity then we chose that one.
				 */
				for(i=0; i<num_forms; i++)
				{
					/* if this i_th formula is in current set of formulae */
					if(cur_world[i] == 1)
					{
						/* add the child of an *ALL* node into the set */
						if(nature[i] == ALL)
						{
							cur_world[childs[i<<1]] = 1;
						}
							
						/* add both childs of an *AND* formula */
						else if(nature[i] == CONJ)
						{
							j = i<<1;
							cur_world[childs[j]] = 1;
							cur_world[childs[j | 1]] = 1;
						}
							
						/* add only one child of the *DISJ* node */
						else if(nature[i] == DISJ)
						{
							j = i<<1;
							if(cur_world[childs[j | 1]] == 0)
							{
								cur_world[childs[j]] = 1;
							}
						}
					} // if
				} // for
					
				isNewWorld = FALSE;
			} // if isNewWorld
			


			/* now we are seeing a saturated wolrd
			 * If the current tableau node is new (is just expanded by 
			 * static rules) then we need to check for any conflict
			 * This is recognized as there is no applications of K-rules yet
			 *
			 * NOTE: This is for the first version. Should have some effective
			 * way to improve this check.
			 */
			if(current_lastPi==num_forms)
			{
				isClosed = FALSE;
				for(i=0; i<num_forms; i++)
				{
					if(cur_world[i] == 1 && nature[i] == NEGA &&
						childs[i << 1] == 0)
						break;
				}
				
				if (i < num_forms) isClosed = TRUE;
				else
				{
					for(i=0; i<num_forms; i++)
					{
						if(cur_world[i] == 1 && nature[i] == PROP)
						{
							for(j=0; j<num_forms; j++)
							{
								if(cur_world[j] == 1 && nature[j] == NEGA
									&& childs[j << 1] == childs[i << 1])
										break;
							}
							
							if(j<num_forms) break;
						}
					}
					
					/* if this model is inconsistent, mark it as close */
					if(i<num_forms)
					{
						isClosed = TRUE;
					}
				}
			}

			
			/* if the current node contains conflicts, move to next
			 * possible sibling
			 */
			if(isClosed == TRUE)
			{
				/* looking for the last application of *OR* rule (not in the 
				 * core from the parent) where we chose the first child. 
				 * There are 2 case:
				 *
				 * If the current world is the root, then we just find the
				 * core is empty. We just simply find the last OR rule with
				 * the first child chosen
				 *
				 * If the current world is not the root, we have to take care
				 * of the core (taken from the parent, due to transitivity)
				 */
				 
				for(i=0; i<num_forms; i++)
				{
					core[i] = 0;
					tmp_world[i] = 0;
				}

				if(stack != null)
				{
					for(i=0; i<num_forms; i++)
					{
						if(nature[i] == ALL && cur_world[i] == 1)
						{ 
							b = (byte)(stack.array[(i>>3)] & (1 << (i&MASK)));
							if(b != 0)
							{
								core[i] = 1;
							}
						}
					}
					/* also the transitional rule to move to this world */
					core[childs[stack.array[ROWLEN] << 1]] = 1;
				}

				
				for(i=num_forms-1; i>=0; i--)
				{
					if(cur_world[i] == 1 && nature[i] == DISJ)
					{
						j = i<<1; /* first child of i */
						k = j | 1; /* second child of i */

						/* if the second child of the node is in current world: need
						 * to erase the child and all of its decendants
						 */
						if(cur_world[childs[k]] == 1) 
						{
							if(core[childs[k]] == 0)
								tmp_world[childs[k]] = 1;
						}

						/* encounter the first *OR* node whose first child is also
						 * in the world. Put its first child to the erase list
						 */
						else if(core[childs[j]] == 0)
						{
							tmp_world[childs[j]] = 1;
							break;
						}
					}
				}

					
				/* if there is a next possible sibling */
				if(i>=0)
				{
					/* erase the formulae that should not occur in the next sibling */
					for(j=0; j<num_forms; j++)
					{
						if(tmp_world[j] == 1)
						{
							/* reset and cur_world at these points */
							cur_world[j] = 0;
							if(nature[j] != PROP && nature[j] != NEGA)
							{
								k = j << 1; /* first child of j */
								if(core[childs[k]] == 0)
									tmp_world[childs[k]] = 1;
								
								k = k | 1; /* second child of j */
								if(childs[k] >= 0 && core[childs[k]] == 0) 
									tmp_world[childs[k]] = 1;
							}
						}
					}


					/* add the second child of the *OR* node above to the sibling */
					cur_world[childs[(i << 1) | 1]] = 1;
						
					/* saturate the sibling */
					for(i++; i<num_forms; i++)
					{
						/* if this i_th formula is in current set of formulae */
						if(cur_world[i] == 1)
						{
							/* add the child of an *ALL* node into the set */
							if(nature[i] == ALL)
							{
								cur_world[childs[i<<1]] = 1;
							}
									
							/* add both childs of an *AND* formula */
							else if(nature[i] == CONJ)
							{
								j = i<<1; /* first child of i */
								cur_world[childs[j]] = 1;
								cur_world[childs[j | 1]] = 1;
							}
									
							/* add the core or first child of the *DISJ* node */
							else if(nature[i] == DISJ)
							{
								j = i<<1; /* first child of i */
								if(core[childs[j | 1]] == 1)
								{
									cur_world[childs[j | 1]] = 1;
								}
								else
								{
									cur_world[childs[j]] = 1;
								}
							}
						} // if
					} // finish saturation

					/* reset last index to the Pi rule */
					current_lastPi = num_forms;
					isCheckedLoop = TRUE;
					continue;
				} // if i >= 0 : next sibling exists

				/* if this is the last possible configuration of this world */
				else
				{
					/* if the current world is root: it is not "good" */
					if(stack == null) return 0;

					/* otherwise, pop the parent node, mark it with closed */
					for(i=0; i<num_forms; i++)
					{						
						b = (byte)(stack.array[(i>>3)] & (1 << (i&MASK)));
						if(b != 0)
						{
							cur_world[i]=1;
						}
						else cur_world[i] = 0;					
					}

					/* when pop some thing from stacks: */
					current_lastPi = stack.array[ROWLEN];
					stack = stack.next;
					isCheckedLoop = TRUE;
					isClosed = TRUE;
					continue;
				}
			} // if isClosed


			/* If this configuration has been checked for conflicts
			 * and it is not closed
			 * try to find the next SOME-formula to apply the K-rule
			 * (Note: we apply K-rules from the bottom up)
			 */
			for(i=current_lastPi-1; i>=0; i--)
			{
				if(nature[i] == SOME && cur_world[i] == 1)
					break;
			}
			
			/* If no more K-rules: every meta-childs is "good", thus this 
			 * is "good". Pop the parent in order to move to its next K-rules
			 */
			if(i < 0)
			{
				/* if the current world is the root, we are done */
				if(stack == null) return 1;
				
				for(i=0; i<num_forms; i++)
				{
					b = (byte)(stack.array[(i>>3)] & (1 << (i&MASK)));
					if(b != 0)
					{
						cur_world[i] = 1;
					}
					else cur_world[i] = 0;
				}

				/* when pop some thing from stacks: */
				current_lastPi = stack.array[ROWLEN];
				stack = stack.next;
				isCheckedLoop = TRUE;

				continue;
			}
			
			/* else take this SOME-formula and apply K-rule to it: new world
			 * set the corresponding last K-formula index
			 */

			
			/* take the child of the Pi-node to the next world */
			for(j=0; j<num_forms; j++) tmp_world[j] = 0;
			tmp_world[childs[i<<1]] = 1;

			/* write the current world into the stack */
			ByteArray new_stack = new ByteArray(new byte[ROWLEN+1], stack);
			stack = new_stack;
			stack.array[ROWLEN] = (byte)i;
			for(i=0; i<num_forms; i++)
			{
				if(cur_world[i] == 1) 
				{
					b = (byte)(stack.array[(i>>3)] | (1<< (i & MASK)));
					stack.array[(i>>3)] = b;
				}
			}
			
			current_lastPi = num_forms;
			
			/* other formulae of the next world: from current ALL formulae */
			for(i=0; i<num_forms; i++)
			{
				if(cur_world[i] == 1 && nature[i] == ALL)
				{
					tmp_world[i] = 1;
				}
			}
			
			/* tmp_world will be current world in next loop. store it into
			 * *cur_world*
			 */
			for(i=0; i<num_forms; i++) cur_world[i] = tmp_world[i];
			isCheckedLoop = FALSE;
			continue;
		}
	}



	public byte simplify()
	{
		int result = 0;
		int i,j,k;
		byte tmp;
		byte num_forms = _num_forms;
		byte childs[] = new byte[num_forms << 1];
		byte nature[] = new byte[num_forms];

		for(i=0; i<num_forms; i++)
		{
			j = i<<1;
			childs[j] = _childs[j];
			childs[j|1] = _childs[j|1];
			nature[i] = _nature[i];
		}
		
		do
		{
			/* scan the parse tree upwards, replace any occurences of
			 * p&~p, p&~1, []~1, <>~1 by ~1
			 * p&p, p&1, p|p, p|~1 by p
			 * p|1, []1 by 1
			 *
			 * This is a linear scan
			 */
			
			result = 0;
			/* scan upwards the tree */
			for(i=num_forms-1; i>=0; i--)
			{
				/* an ALL node: propogate the TRUE and NOT TRUE */
				if(nature[i] == card.proverInterface.ALL)
				{
					j = i<<1;
					if(nature[childs[j]] == card.proverInterface.PROP &&
						childs[childs[j] << 1] == 0)
					{
						nature[i] = card.proverInterface.PROP;
						childs[j] = 0;
						result++;
					}
					
					else if(nature[childs[j]] == card.proverInterface.NEGA
						&& childs[childs[j] << 1] == 0)
					{
						nature[i] = card.proverInterface.NEGA;
						childs[j] = 0;
						result++;
					}
				}
				
				/* propogate the NOT TRUE for SOME: similar */
				else if(nature[i] == card.proverInterface.SOME)
				{
					j = i<<1;
					if(nature[childs[j]] == card.proverInterface.NEGA
						&& childs[childs[j] << 1] == 0)
					{
						nature[i] = card.proverInterface.NEGA;
						childs[j] = 0;
						result++;
					}
				}
				
				/* CONJ node:
				 * p&p => p, ~p&~p => ~p, 1&p => p, ~1&p => ~1
				 */
				else if(nature[i] == card.proverInterface.CONJ)
				{
					j = i<<1;
					k = j | 1;
					/* if they are the same, then replace this node by one of
					 * its childs. This is applied only for PROP and NEGA
					 */
					if(nature[childs[j]] == nature[childs[k]] &&
						childs[childs[j] << 1] == childs[childs[k] << 1])
					{
						nature[i] = nature[childs[j]];
						childs[k] = childs[(childs[j] << 1) | 1];
						childs[j] = childs[childs[j] << 1];
						result++;
					}
					
					/* if one of them is NEGA 0: replace by NEGA 0 */
					else if((nature[childs[j]] == card.proverInterface.NEGA &&
						childs[childs[j] << 1] == 0) ||
						(nature[childs[k]] == card.proverInterface.NEGA &&
						childs[childs[k] << 1] == 0))
					{
						nature[i] = card.proverInterface.NEGA;
						childs[j] = 0;
						result++;
					}
						
					/* if one of them is PROP 0: replace by the other child */
					else if(nature[childs[j]] == card.proverInterface.PROP &&
						childs[childs[j] << 1] == 0)
					{
						nature[i] = nature[childs[k]];
						childs[j] = childs[childs[k] << 1];
						childs[k] = childs[(childs[k] << 1) | 1];
						
						result ++;
					}
	
					else if(nature[childs[k]] == card.proverInterface.PROP &&
						childs[childs[k] << 1] == 0)
					{
						nature[i] = nature[childs[j]];
						childs[k] = childs[(childs[j] << 1) | 1];
						childs[j] = childs[childs[j] << 1];
						
						result ++;
					}
				
					/* if they are negation of each other */	
					else if((childs[childs[j] << 1] == childs[childs[k] << 1]) &&
						((nature[childs[j]] == card.proverInterface.PROP &&
						  nature[childs[k]] == card.proverInterface.NEGA) ||
						(nature[childs[j]] == card.proverInterface.NEGA &&
						  nature[childs[k]] == card.proverInterface.PROP)))
						
					{
						nature[i] = card.proverInterface.NEGA;
						childs[j] = 0;
						
						result ++;
					}
				}
							
				/* DISJ node in the parse tree
				 * p|p => p, ~p|~p => ~p, 1|p => 1, ~1|p => p
				 */
				else if(nature[i] == card.proverInterface.DISJ)
				{
					j = i<<1;
					k = j|1;
					/* if they are the same, then replace this node by one of
					 * its childs. This is applied only for PROP and NEGA
					 */
					if(nature[childs[j]] == nature[childs[k]] &&
						childs[childs[j] << 1] == childs[childs[k] << 1])
					{
						nature[i] = nature[childs[j]];
						childs[j] = childs[childs[j] << 1];
						childs[k] = childs[(childs[j] << 1) | 1];
						result++;
					}
	
					/* if one of them is PROP 0: replace by PROP 0 */
					else if(nature[childs[j]] == card.proverInterface.PROP &&
						childs[childs[j]<<1] == 0)
					{
						nature[i] = card.proverInterface.PROP;
						childs[j] = 0;
						
						result ++;
					}
	
					else if(nature[childs[k]] == card.proverInterface.PROP &&
						childs[childs[k] << 1] == 0)
					{
						nature[i] = card.proverInterface.PROP;
						childs[j] = 0;
						
						result ++;
					}
	
					/* if one of them is NEGA 0: replace by the other childs */
					else if(nature[childs[j]] == card.proverInterface.NEGA &&
						childs[childs[j] << 1] == 0)
						
					{
						nature[i] = nature[childs[k]];
						childs[j] = childs[childs[k] << 1];
						childs[k] = childs[(childs[k] << 1) | 1];
						
						result ++;
					}
					
					else if(nature[childs[k]] == card.proverInterface.NEGA &&
						childs[childs[k] << 1] == 0)
					{
						nature[i] = nature[childs[j]];
						childs[k] = childs[(childs[j] << 1) | 1];
						childs[j] = childs[childs[j] << 1];
						
						result ++;
					}
				}
			}
	

			/* mark each DISJ and SOME node as bad if it contains -1 of both
			 * p and ~p. These should be "direct" descendants, i.e there is
			 * not any SOME formulae in between them in the parse tree
			 */

			/* find the lowest DISJ or SOME ancestor of every node */
			byte ances[] = new byte [num_forms];
			for(i=0; i<num_forms; i++) ances[i] = -1;
			
			ances[0] = 0;

			for(i=0; i<num_forms; i++)
			{
				j = i<<1;
				if(nature[i] == card.proverInterface.DISJ)
				{
					k = (j|1);
					ances[childs[j]] = childs[j];
					ances[childs[k]] = childs[k];
				}
				
				else if(nature[i] == card.proverInterface.SOME)
				{
					ances[childs[j]] = (byte)i;
				}

				else if(nature[i] == card.proverInterface.ALL)
				{
					ances[childs[j]] = ances[i];
				}

				else if(nature[i] == card.proverInterface.CONJ)
				{
					k = (j|1);
					ances[childs[j]] = ances[i];
					ances[childs[k]] = ances[i];
				}
			}
	
	
			/* scan the tree upwards, if NEGA TRUE has ances *an* or both p and ~p
			 * have ances *an* then replace *an* by NEGA TRUE
			 */
			for(i=num_forms-1; i>= 0; i--)
			{
				if(ances[i] < 0) continue;
				j = (i<<1);
				if(nature[i] == card.proverInterface.NEGA &&
					childs[j] == 0)
				{
					j = (ances[i]<<1);
					nature[ances[i]] = card.proverInterface.NEGA;
					childs[j] = 0;
					childs[j|1] = -1;
					continue;
				}
				
				if(nature[i] == card.proverInterface.PROP)
				{
					for(int l=(i-1); l>=0; l--)
					{
						if(nature[l] == card.proverInterface.NEGA && ances[l] == ances[i])
						{
							k = (l<<1);
							if(childs[k] == childs[j])
							{
								j = (ances[i]<<1);
								nature[ances[i]] = card.proverInterface.NEGA;
								childs[j] = 0;
								childs[j|1] = -1;
								continue;
							}
						}
					}
				}
	
				else if(nature[i] == card.proverInterface.NEGA)
				{
					for(int l=(i-1); l>=0; l--)
					{
						if(nature[l] == card.proverInterface.PROP && ances[l] == ances[i])
						{
							k = (l<<1);
							if(childs[k] == childs[j])
							{
								j = (ances[i]<<1);
								nature[ances[i]] = card.proverInterface.NEGA;
								childs[j] = 0;
								childs[j|1] = -1;
								continue;
							}
						}
					}
				}
			}
	
			/* reorganize */
			byte[] stack = new byte[num_forms];
			byte stackpt = 0;
			byte num_new = 1;
			byte childs_tmp[] = new byte [num_forms << 1];
			byte nature_tmp[] = new byte[num_forms];

			stack[stackpt] = 0; /* the root */
			while (stackpt < num_new)
			{
				/* pop from stack */
				nature_tmp[stackpt] = nature[stack[stackpt]];
				if(nature_tmp[stackpt] == card.proverInterface.PROP ||
					nature_tmp[stackpt] == card.proverInterface.NEGA)
				{
					childs_tmp[stackpt << 1] = childs[stack[stackpt] << 1];
					childs_tmp[(stackpt << 1) | 1] = -1;
				}
				
				else
				{
					/* push the (first) child into stack */
					childs_tmp[stackpt << 1] = num_new;
					stack[num_new] = childs[stack[stackpt] << 1];
					num_new++;
	
					/* if binary: push the second child into stack */				
					if(nature[stack[stackpt]] == card.proverInterface.DISJ ||
						nature[stack[stackpt]] == card.proverInterface.CONJ)
					{
						childs_tmp[(stackpt << 1) | 1] = num_new;
						stack[num_new] = childs[(stack[stackpt] << 1) | 1];
						num_new ++;
					}
					
					else childs_tmp[(stackpt << 1) | 1] = -1;
				}
				stackpt ++;
			}
	
			result = (byte)(num_forms - stackpt);

			/* copy to persistent fields */			
			if(result > 0)
			{
				num_forms = stackpt;
				childs = new byte[num_forms << 1];
				nature = new byte[num_forms];

				for(i = 0; i<num_forms; i++)
				{
					j = i<<1;
					nature[i] = nature_tmp[i];
					childs[j] = childs_tmp[j];
					childs[j|1] = childs_tmp[j|1];
				}
			}
		} while(result > 0);

		byte res = (byte)(_num_forms - num_forms);
		if(res > 0)
		{
			for(i=0; i<num_forms; i++)
			{
				j = i<<1;
				_nature[i] = nature[i];
				_childs[j] = childs[j];
				_childs[j|1] = childs[j|1];
			}
			_num_forms = num_forms;
		}
		return res;
	}
}